"""
This module contains the data model for the Takeoff SDK. This is used to parse the environment variables
"""
# ───────────────────────────────────────────────────── imports ────────────────────────────────────────────────────── #

from typing import Dict, Optional

from pydantic import BaseModel, ConfigDict, Field, validator

# ──────────────────────────────────────────────────────────────────────────────────────────────────────────────────── #
#                                               Configuration Data Model                                               #
# ──────────────────────────────────────────────────────────────────────────────────────────────────────────────────── #

__all__ = ["TakeoffEnvSetting"]


class TakeoffEnvSetting(BaseModel):
    """
    This class contains the data model for the Takeoff SDK. This setting is used to set the environment variables
    """

    model_config = ConfigDict(extra="forbid", protected_namespaces=())

    # ------------------------------------- model configuration -------------------------------------- #

    # NOTE: these are expected to be set by the user
    model_name: str = Field(..., env="TAKEOFF_MODEL_NAME")
    device: str = Field(..., env="TAKEOFF_DEVICE")

    backend: Optional[str] = Field(None, env="TAKEOFF_BACKEND")

    max_batch_size: int = Field(8, env="TAKEOFF_MAX_BATCH_SIZE")
    batch_duration_millis: int = Field(100, env="TAKEOFF_BATCH_DURATION_MILLIS")
    access_token: Optional[str] = Field(None, env="TAKEOFF_ACCESS_TOKEN")
    cuda_visible_devices: Optional[str] = Field(None, env="TAKEOFF_CUDA_VISIBLE_DEVICES")

    refill_threshold: float = Field(0.5, env="TAKEOFF_REILL_THRESHOLD")
    disable_continuous_generation: bool = Field(False, env="TAKEOFF_DISABLE_CONTINUOUS_GENERATION")
    consumer_group: str = Field("primary", env="TAKEOFF_CONSUMER_GROUP")

    redis_host: str = Field("localhost", env="TAKEOFF_REDIS_HOST")
    redis_port: int = Field(6379, env="TAKEOFF_REDIS_PORT")

    heartbeat_wait_interval: int = Field(10, env="TAKEOFF_HEARTBEAT_WAIT_INTERVAL")

    # JF Config
    nvlink_unavailable: int = Field(0, env="TAKEOFF_NVLINK_UNAVAILABLE")

    tensor_parallel: int = Field(1, env="TAKEOFF_TENSOR_PARALLEL")
    disable_cuda_graph: int = Field(0, env="TAKEOFF_DISABLE_CUDA_GRAPH")
    quant_type: str = Field("auto", env="TAKEOFF_QUANT_TYPE")  # auto, awq
    max_seq_len: int = Field(128, env="TAKEOFF_MAX_SEQUENCE_LENGTH")
    disable_static: int = Field(0, env="TAKEOFF_DISABLE_STATIC")

    export_traces: bool = Field(False, env="TAKEOFF_EXPORT_TRACES")
    traces_host: str = Field("http://localhost:4317", env="TAKEOFF_TRACES_HOST")
    metrics_target: str = Field("redis", env="TAKEOFF_METRICS_TARGET")

    metrics_log_level: str = Field("INFO", env="TAKEOFF_METRICS_LOG_LEVEL")

    is_echo: bool = Field(False, env="TAKEOFF_ECHO")

    log_level: str = Field("INFO", env="TAKEOFF_LOG_LEVEL")
    rust_log_level: str = Field("INFO", env="RUST_LOG")

    run_name: str = Field("takeoff_default", env="TAKEOFF_RUN_NAME")

    license_key: Optional[str] = Field(None, env="LICENSE_KEY")

    @validator("device")
    def validate_device(cls, v):
        if v not in ["cpu", "cuda"]:
            raise ValueError("device must be either cpu or cuda")
        return v

    def settings_to_env_vars(self) -> Dict[str, str]:
        env_var_mapping = {}
        for field_name, model_field in self.__dict__.items():
            field_info = self.model_fields[field_name]
            default_value = field_info.default

            # Only add to env_var_mapping if the current value is different from the default value
            if model_field != default_value:
                if field_info.json_schema_extra is not None and "env" in field_info.json_schema_extra:
                    env_var_name = field_info.json_schema_extra["env"]

                    # Check if the value is of type bool and convert to lowercase string
                    value = model_field
                    if isinstance(value, bool):
                        value = str(value).lower()

                    env_var_mapping[env_var_name] = value

        return env_var_mapping



"""This script is used to bump the version of the package. pyproject.toml is used to store the version."""
import sys
import toml


def read_version():
    with open("pyproject.toml", "r") as file:
        data = toml.load(file)
    return data["project"]["version"]


def write_version(new_version):
    # Read the content of the file
    with open("pyproject.toml", "r") as file:
        content = file.readlines()

    # Find and update the version line
    for i, line in enumerate(content):
        if line.startswith("version = "):
            content[i] = f'version = "{new_version}"\n'
            break

    # Write the updated content back to the file
    with open("pyproject.toml", "w") as file:
        file.writelines(content)


def bump_version(version_part):
    major, minor, patch = map(int, read_version().split("."))
    if version_part == "major":
        major += 1
        minor = 0
        patch = 0
    elif version_part == "minor":
        minor += 1
        patch = 0
    elif version_part == "patch":
        patch += 1
    return f"{major}.{minor}.{patch}"


if __name__ == "__main__":
    part = sys.argv[1] if len(sys.argv) > 1 else "patch"
    new_version = bump_version(part)
    write_version(new_version)
    print(f"Version updated to: {new_version}")



from unittest.mock import MagicMock, patch

from takeoff_sdk import Takeoff, TakeoffEnvSetting


def test_takeoff_initialization():
    model_name = "test_model"
    device = "cpu"
    takeoff = Takeoff(model_name=model_name, device=device)

    assert takeoff.model_name == model_name
    assert takeoff.device == device
    assert isinstance(takeoff.takeoff_config, TakeoffEnvSetting)
    assert takeoff.models == []
    assert takeoff.consumer_groups == {}
    assert takeoff.server_url is None
    assert takeoff.management_url is None
    assert takeoff.container is None


def test_takeoff_from_config():
    config = TakeoffEnvSetting(model_name="test_model", device="cpu")
    takeoff = Takeoff.from_config(config)

    assert takeoff.model_name == config.model_name
    assert takeoff.device == config.device


@patch("takeoff_sdk.sdk.takeoff.start_takeoff", return_value=(8000, 8001, MagicMock()))
@patch("takeoff_sdk.sdk.takeoff.is_takeoff_loading", side_effect=[True, True, False])
def test_takeoff_start(mock_is_loading, mock_start_takeoff):
    takeoff = Takeoff(model_name="test_model", device="cpu")

    # Assuming readers() method returns some mock data
    mock_readers = {"group1": [{"model_name": "model1"}]}
    takeoff.readers = MagicMock(return_value=mock_readers)

    takeoff.start()

    assert takeoff.server_url == "http://localhost:8000"
    assert takeoff.management_url == "http://localhost:8001"
    assert "model1" in takeoff.consumer_groups
    assert "model1" in takeoff.models
    mock_start_takeoff.assert_called_once_with(takeoff.takeoff_config)
    assert mock_is_loading.call_count == 3



"""Takeoff Object SDK"""
# ───────────────────────────────────────────────────── imports ────────────────────────────────────────────────────── #

import time

import requests
from loguru import logger

from .schema import TakeoffEnvSetting
from .utils import is_takeoff_loading, start_takeoff, is_docker_logs_error

# ──────────────────────────────────────────────────────────────────────────────────────────────────────────────────── #
#                                                       Takeoff                                                        #
# ──────────────────────────────────────────────────────────────────────────────────────────────────────────────────── #


class Takeoff:
    def __init__(self, model_name, device, **kwargs):
        self.model_name = model_name
        self.device = device

        self.models = []
        self.consumer_groups = {}
        self.server_url = None
        self.management_url = None
        self.openai_url = None
        self.container = None

        self.takeoff_config = TakeoffEnvSetting(model_name=model_name, device=device, **kwargs)

    def cleanup(self):
        if self.container:
            self.container.remove(force=True)

    @classmethod
    def from_config(cls, takeoff_config: TakeoffEnvSetting):
        takeoff_args = takeoff_config.model_dump()
        return cls(**takeoff_args)

    def start(self):
        """Start the Takeoff server and add the model to the primary consumer group"""

        logger.info("Starting Takeoff server...")
        self.takeoff_port, self.management_port, self.openai_port, self.container = start_takeoff(self.takeoff_config)

        self.server_url = f"http://localhost:{self.takeoff_port}"
        self.management_url = f"http://localhost:{self.management_port}"
        self.openai_url = f"http://localhost:{self.openai_port}"

        logger.info(f"Takeoff server running at: {self.server_url}")
        logger.info(f"Takeoff management server running at: {self.management_url}")
        logger.info(f"Takeoff openai compatible server running at: {self.openai_url}")

        for _ in range(100):  # Some models take a while to download
            if not is_takeoff_loading(self.server_url):
                break

            traceback = is_docker_logs_error(self.container)

            if traceback is not None:
                logger.error("Takeoff server failed to start. Error in docker logs. Cleaning up...")
                self.cleanup()
                raise Exception(
                    "Takeoff server failed to start due to error in docker logs. See below for details. \n" + traceback
                )
            logger.info("building...")
            time.sleep(3)
        else:
            raise Exception("Takeoff server build timed out")

        logger.info("server ready!")

        readers = self.readers()
        for group, details in readers.items():
            model_name = details[0]["model_name"]  # Assume one model for each group for now
            self.consumer_groups[model_name] = group
            self.models.append(model_name)

    def add_model(self, model_name: str, backend: str, device: str, consumer_group: str = "primary"):
        """Spawn a new reader for the given model and add it to the given consumer group

        Args:
            model_name (str): model name
            backend (str): backend
            device (str): device
            consumer_group (str, optional): consumer group. Defaults to "primary".
        """
        response = requests.post(
            self.management_url + "/reader",
            json={"model_name": model_name, "device": device, "backend": backend, "consumer_group": consumer_group},
        )
        assert response.ok
        self.models.append(model_name)
        self.consumer_groups[model_name] = consumer_group
        logger.info(f"Added model {model_name} to consumer group {consumer_group}")

    def readers(self):
        response = requests.get(self.management_url + "/reader_groups")
        return response.json()



"""Utility functions for Takeoff SDK"""
# ───────────────────────────────────────────────────── imports ────────────────────────────────────────────────────── #

import os
import re
import uuid
from pathlib import Path

import docker
import requests
from .schema import TakeoffEnvSetting


def start_container(
    image_name: str,
    container_name: str,
    ports: dict,
    volumes: list,
    environment_vars: dict,
    device_requests: list,
    shm_size: str,
):
    """Start a docker container.

    Args:
        image_name (str): image name with tag. example: "takeoff:cpu"
        container_name (str): container name
        ports (dict): port mapping. example: {"3000/tcp": None, "3001/tcp": None}
        volumes (list): volume mapping. example: [path1:path2]
        environment_vars (dict): environment variables. example: {"TAKEOFF_MAX_BATCH_SIZE": 1}
        device_requests (list): device requests. example: [docker.types.DeviceRequest(count=-1, capabilities=[["gpu"]])]
        shm_size (str): shm size. example: "4G"
    """
    client = docker.from_env()

    container = client.containers.run(
        image_name,
        detach=True,
        auto_remove=True,
        environment=environment_vars,
        name=container_name,
        device_requests=device_requests,
        volumes=volumes,
        ports=ports,
        shm_size=shm_size,
    )
    return container


def is_docker_logs_error(container):
    """Check the docker logs for errors

    Args:
        container (docker.models.containers.Container): docker container

    Returns:
        str: traceback if error, None if no error
    """

    # Stream logs from the container
    try:
        logs = container.logs().decode("utf-8")

        # Regular expression to match tracebacks
        traceback_pattern = r"Traceback \(most recent call last\):.*?(?=\d{4}-\d{2}-\d{2}T\d{2}:\d{2}:\d{2}\.\d{6}Z|$)"
        tracebacks = re.findall(traceback_pattern, logs, re.DOTALL)

        if len(tracebacks) > 0:
            print("=====================================")
            print("Error while starting Takeoff server!")

            # print(f"Tracebacks: {tracebacks[0]}")
            return tracebacks[0]

    except Exception as e:
        print(f"Error while reading logs: {e}")
        return f"Error while reading logs: {e}"

    return None


def is_takeoff_loading(server_url: str) -> bool:
    """Check if the Takeoff server is loading

    Args:
        server_url (str): server url

    Returns:
        bool: True if the server is loading, False if it is ready
    """
    try:
        response = requests.get(server_url + "/healthz")
        return not response.ok
    except requests.exceptions.ConnectionError:
        return True


def start_takeoff(takeoff_config: TakeoffEnvSetting) -> tuple:
    """Start a Takeoff server using the given configuration

    Args:
        takeoff_config (TakeoffEnvSetting): Takeoff configuration

    Returns:
        tuple: (takeoff_port, management_port, container)
    """

    image = os.environ.get("TAKEOFF_IMAGE", "tytn/takeoff-pro:0.8.0-gpu")

    random_name = "takeoff_" + str(uuid.uuid4())[:8]
    volumes = [f"{Path.home()}/.iris_cache:/code/models"]
    device_requests = (
        [docker.types.DeviceRequest(count=-1, capabilities=[["gpu"]])] if takeoff_config.device == "cuda" else None
    )
    environtment_vars = takeoff_config.settings_to_env_vars()

    container = start_container(
        image_name=image,
        container_name=random_name,
        ports={"3000/tcp": None, "3001/tcp": None, "3003/tcp": None},
        volumes=volumes,
        environment_vars=environtment_vars,
        device_requests=device_requests,
        shm_size="4G",
    )

    container.reload()
    takeoff_port = container.ports["3000/tcp"][0]["HostPort"]
    management_port = container.ports["3001/tcp"][0]["HostPort"]
    openai_port = container.ports["3003/tcp"][0]["HostPort"]

    return takeoff_port, management_port, openai_port, container



"""Takeoff CLI for starting and managing Takeoff servers."""
# ───────────────────────────────────────────────────── imports ────────────────────────────────────────────────────── #

import docker
import typer

from .sdk.takeoff import Takeoff

main = typer.Typer()

# Create a Docker client
client = docker.from_env()


@main.command()
def start(
    model_name: str = typer.Option(..., "--model", "-m", help="The models to optimize."),
    backend: str = typer.Option(..., "--backend", "-b", help="The backend to use."),
    device: str = typer.Option(..., "--device", "-d", help="The device to use."),
):
    try:
        takeoff = Takeoff(model_name=model_name, backend=backend, device=device)
        takeoff.start()
    except Exception as e:
        typer.echo(f"Server start failed: {e}")


@main.command()
def hello(name: str):
    typer.echo(f"Hello {name}")


@main.command()
def list():
    typer.echo("Listing running takeoff servers...")

    # List all running containers
    containers = client.containers.list()

    # Print the list of running containers
    typer.echo("Running containers:")
    for container in containers:
        if "takeoff" in container.name:
            typer.echo(f"ID: {container.short_id}, Name: {container.name}")


@main.command()
def down(
    id: str = typer.Option(None, "--id", "-i", help="The ID of the server to shut down."),
    is_all: bool = typer.Option(False, "--all", "-a", help="Shut down all running servers."),
):
    if id is None and not is_all:
        typer.echo("Please specify the ID of the server to shut down. Or use --all to shut down all running servers.")
        return

    typer.echo("Shutting down server...")
    if id is not None:
        typer.echo(f"Shutting down server with ID {id}...")
        try:
            client.containers.get(id).stop()
            client.containers.get(id).remove()
            typer.echo("Server shut down")
        except Exception as e:
            typer.echo(f"Server shut down failed: {e}")

    if is_all:
        typer.echo("Shutting down all servers...")
        # List all running containers
        containers = client.containers.list()

        # Print the list of running containers
        for container in containers:
            if "takeoff" in container.name:
                typer.echo(f"Shutting down -> ID: {container.short_id}, Name: {container.name}")

                client.containers.get(container.name).stop()
                client.containers.get(container.name).remove()

        typer.echo("Server shut down")


if __name__ == "__main__":
    main()



import pytest
from pydantic import ValidationError
from takeoff_sdk import TakeoffEnvSetting


def test_data_model_with_default_values():
    # Test with default values
    settings = TakeoffEnvSetting(model_name="default_model", device="cpu")
    assert settings.model_name == "default_model"
    assert settings.device == "cpu"
    assert settings.backend is None
    assert settings.max_batch_size == 8
    assert settings.batch_duration_millis == 100
    assert settings.access_token is None
    assert settings.cuda_visible_devices is None
    assert not settings.disable_continuous_generation
    assert settings.consumer_group == "primary"
    assert settings.redis_host == "localhost"
    assert settings.redis_port == 6379
    assert settings.heartbeat_wait_interval == 10
    assert settings.nvlink_unavailable == 0
    assert settings.tensor_parallel == 1
    assert settings.disable_cuda_graph == 0
    assert settings.quant_type == "auto"
    assert settings.max_seq_len == 128
    assert settings.disable_static == 0
    assert not settings.export_traces
    assert settings.traces_host == "http://localhost:4317"
    assert settings.metrics_target == "redis"
    assert settings.metrics_log_level == "INFO"
    assert not settings.is_echo
    assert settings.log_level == "INFO"
    assert settings.rust_log_level == "INFO"
    assert settings.run_name == "takeoff_default"
    assert settings.license_key is None


def test_settings_to_env_vars_with_custom_values():
    # Test with custom values
    settings = TakeoffEnvSetting(model_name="custom_model", device="cpu", max_batch_size=16)
    env_vars = settings.settings_to_env_vars()
    expected_vars = {"TAKEOFF_MODEL_NAME": "custom_model", "TAKEOFF_DEVICE": "cpu", "TAKEOFF_MAX_BATCH_SIZE": 16}
    assert env_vars == expected_vars


def test_settings_to_env_vars_with_bool_values():
    # Test with bool values, it shoud be converted to lower case string
    settings = TakeoffEnvSetting(model_name="custom_model", device="cpu", disable_continuous_generation=True)
    env_vars = settings.settings_to_env_vars()
    expected_vars = {
        "TAKEOFF_MODEL_NAME": "custom_model",
        "TAKEOFF_DEVICE": "cpu",
        "TAKEOFF_DISABLE_CONTINUOUS_GENERATION": "true",
    }
    assert env_vars == expected_vars


def test_validate_device():
    # Valid devices should not raise an error
    valid_devices = ["cpu", "cuda"]
    for device in valid_devices:
        settings = TakeoffEnvSetting(model_name="test_model", device=device)
        assert settings.device == device

    # Invalid device should raise an error
    with pytest.raises(ValueError) as e:
        TakeoffEnvSetting(model_name="test_model", device="invalid_device")
    assert "device must be either cpu or cuda" in str(e.value)


def test_invalid_attribute():
    # Attempting to create an instance with an undefined attribute should raise an error
    with pytest.raises(ValidationError):
        TakeoffEnvSetting(model_name="test_model", device="cpu", undefined_attribute="value")
